local Macronomy_EventFrame, events = CreateFrame("FRAME"), {}
local MacronomyAddonIsLoaded = false
local MacronomyMacrosAreLoaded = false
local MacronomyEnteredWorld = false
local MacronomyTimer = 0
local MacronomyInterval = .1
local MacronomyButtonUpdate = false
local MacronomyClassList = {}
local MacronomyExportClassList = {}
local MacronomySpecList = {}
local MacronomyExportSpecList = {}
local MacronomyButtonList = {}
local MacronomyExportButtonList = {}
local MacronomyCategoryList = {"Default"}
local MMY_locale = GetLocale()
local MMY_InScrollFrame = false
--------------------------------------------------------------------
function events:ADDON_LOADED(...)
	local addonName = ...
	if addonName == "Macronomy" then
		MacronomyAddonIsLoaded = true
		Macronomy_Init()
		Macronomy_EventFrame:UnregisterEvent("ADDON_LOADED")
	end
end

function events:UPDATE_MACROS(...)
	MacronomyMacrosAreLoaded = true
	Macronomy_Init()
	Macronomy_EventFrame:UnregisterEvent("UPDATE_MACROS")
end

function events:PLAYER_ENTERING_WORLD(...)
	MacronomyEnteredWorld = true
	Macronomy_Init()
	Macronomy_EventFrame:UnregisterEvent("PLAYER_ENTERING_WORLD")
end

function events:ACTIVE_TALENT_GROUP_CHANGED(...)
	MacronomyButtonUpdate = true
end

function events:BAG_UPDATE(...)
	MacronomyButtonUpdate = true
end

function events:SPELLS_CHANGED(...)
	MacronomyButtonUpdate = true
end

function events:UNIT_INVENTORY_CHANGED(unit)
	if unit == "player" then MacronomyButtonUpdate = true end
end

--------------------------------------------------------------------
Macronomy_EventFrame:SetScript("OnEvent", function(self, event, ...)
	events[event](self, ...)
end)

Macronomy_EventFrame:SetScript("OnUpdate", function(self, elapsed)
	MacronomyTimer = MacronomyTimer + elapsed
	if MacronomyTimer > MacronomyInterval then
		Macronomy_SetButtons()
		MacronomyTimer = 0
	end
end)
--------------------------------------------------------------------
for i, v in pairs(events) do
	Macronomy_EventFrame:RegisterEvent(i)
end
--------------------------------------------------------------------
function MMY_echo(txt, override)
	if Macronomy ~= nil and Macronomy.troubleshootMode or override then
		if txt ~= nil and type(txt) == "string" then
			DEFAULT_CHAT_FRAME:AddMessage(txt)
		else
			DEFAULT_CHAT_FRAME:AddMessage("MMY: Supplied variable is nil")
		end
	end
end
--------------------------------------------------------------------
function Macronomy_Init()
	if MacronomyAddonIsLoaded and MacronomyMacrosAreLoaded and MacronomyEnteredWorld then
		local L = MacronomyLocalizations
		PopulateLists()

		MacronomyFrame:Hide()
		MacronomyFrame.name = "Macronomy"
		InterfaceOptions_AddCategory(MacronomyFrame);


		if not MacronomyByChar then MacronomyByChar = {} end
		MacronomyByChar.currentClass = UnitClass("player")
		MacronomyByChar.currentSpec = Macronomy_GetPlayerSpec()
		MacronomyByChar.currentCategory = "Default"
		MacronomyByChar.currentButton = "m1"
		if not MacronomyByChar.categoryInUse then MacronomyByChar.categoryInUse = 1 end

		--MMY_OldInitMacronomyVar()
		MMY_InitMacronomyVar()
		
		MacronomyMacroFrame_ebxMacro:SetTextInsets(8,5,5,16)
		MacronomyNewCategoryFrame_ebxNewCat:SetTextInsets(8,5,5,5)
		MacronomyExportFrame_scrollFrame_ebxExport:SetTextInsets(8,5,5,5)
		MacronomyImportFrame_scrollFrame_ebxImport:SetTextInsets(8,5,5,5)
		Macronomy_SetDropDowns()
		MacronomyOptionsFrame_chkShowTTByDefault:SetChecked(Macronomy.addShowTooltipByDefault)
		MMY_setDisableTTVisibleState()
		MacronomyMacroFrame_btnSave:Disable()
		MacronomyButtonUpdate = true
		MacronomyFrame:RegisterForDrag("LeftButton")
		Macronomy.active = false
		MMY_echo("Macronomy initialized", true)
	end
end
--------------------------------------------------------------------
function PopulateLists()
	-- Populate class and spec lists
	for i=1,GetNumClasses() do
		local currentClass = GetClassInfoByID(i)
		table.insert(MacronomyClassList, currentClass)

		MacronomySpecList[currentClass] = {}
		for j=1,GetNumSpecializationsForClassID(i) do
			local specID, specName = GetSpecializationInfoForClassID(i,j)
			table.insert(MacronomySpecList[currentClass], specName)
		end
		sort(MacronomySpecList[MacronomyClassList[i]])
		table.insert(MacronomySpecList[currentClass], "NoSpec")
	end
	sort(MacronomyClassList)

	-- Populate export class list
	MacronomyExportClassList = MMY_deepCopy(MacronomyClassList)
	table.insert(MacronomyExportClassList, 1, "All")

	-- Populate export spec list
	MacronomyExportSpecList["All"] = {"All"}
	for i=1,getn(MacronomyClassList) do
		local currentClass = MacronomyClassList[i]
		MacronomyExportSpecList[currentClass] = {"All"}
		for j=1,getn(MacronomySpecList[currentClass]) do
			table.insert(MacronomyExportSpecList[currentClass], MacronomySpecList[currentClass][j])
		end
	end

	-- Populate button lists
	for i=1,9 do
		table.insert(MacronomyButtonList, "m" .. i)
	end
	table.insert(MacronomyButtonList, "m0")
	table.insert(MacronomyButtonList, "m-")
	table.insert(MacronomyButtonList, "m=")
	MacronomyExportButtonList = MMY_deepCopy(MacronomyButtonList)
	table.insert(MacronomyExportButtonList, 1, "All")

	MacronomyCategoryList[1] = "Default"
end
--------------------------------------------------------------------
function MMY_InitMacronomyVar()
	local L = MacronomyLocalizations

	local populateNewFormatVar = function()
		Macronomy.enUS = {}
		Macronomy[MMY_locale] = {}
		for classNum = 1,  getn(MacronomyClassList) do
			local class4Init = MacronomyClassList[classNum]
			Macronomy[MMY_locale][class4Init] = {}

			for specNum = 1, getn(MacronomySpecList[MacronomyClassList[classNum]]) do
				local spec4Init = MacronomySpecList[MacronomyClassList[classNum]][specNum]
				Macronomy[MMY_locale][class4Init][spec4Init] = {}
				Macronomy[MMY_locale][class4Init][spec4Init].categories = {}
				Macronomy_InitNewCategory(class4Init, spec4Init, "Default")
			end
		end
		Macronomy.troubleshootMode = false
	end

	if not Macronomy then
		Macronomy = {}
		populateNewFormatVar()
		MMY_echo("Macronomy variable initialized", true)
	elseif not Macronomy.enUS then
		populateNewFormatVar()

		-- Copy old Macronomy fields to new field location
		for classNum=1,getn(MacronomyClassList) do
			local class4copy = MacronomyClassList[classNum]
			for specNum=1,getn(MacronomySpecList[class4copy]) do
				local spec4copy = MacronomySpecList[class4copy][specNum]
				for catNum=1,getn(Macronomy[class4copy][spec4copy].categories) do
					local cat4copy = Macronomy[class4copy][spec4copy].categories[catNum]
					if cat4copy ~= "Default" then table.insert(Macronomy[MMY_locale][class4copy][spec4copy].categories, cat4copy) end
					Macronomy[MMY_locale][class4copy][spec4copy][cat4copy] = {}
					for btnNum = 1, 12 do
						Macronomy[MMY_locale][class4copy][spec4copy][cat4copy][MacronomyButtonList[btnNum]] = {}
						Macronomy[MMY_locale][class4copy][spec4copy][cat4copy][MacronomyButtonList[btnNum]].btn = Macronomy[class4copy][spec4copy][cat4copy][MacronomyButtonList[btnNum]]
						Macronomy[MMY_locale][class4copy][spec4copy][cat4copy][MacronomyButtonList[btnNum]].autoShowTooltipDisabled = false
					end
				end
			end
		end

		MMY_echo("Macronomy variable updated to new format", true)
	end
end

function MMY_OldInitMacronomyVar()
	--This section creates the Macronomy variable in the old foramt
	if not Macronomy then
		Macronomy = {}
		for classNum = 1, getn(MacronomyClassList) do
			Macronomy[MacronomyClassList[classNum]] = {}
		
			for specNum = 1, getn(MacronomySpecList[MacronomyClassList[classNum]]) do
				Macronomy[MacronomyClassList[classNum]][MacronomySpecList[MacronomyClassList[classNum]][specNum]] = {}
				Macronomy[MacronomyClassList[classNum]][MacronomySpecList[MacronomyClassList[classNum]][specNum]].categories = {"Default"}
				Macronomy_InitCategory(MacronomyClassList[classNum], MacronomySpecList[MacronomyClassList[classNum]][specNum], "Default")
			end
		end
	end
end
--------------------------------------------------------------------
function MMY_deepCopy(orig)
    local orig_type = type(orig)
    local copy
    if orig_type == 'table' then
        copy = {}
        for orig_key, orig_value in next, orig, nil do
            copy[MMY_deepCopy(orig_key)] = MMY_deepCopy(orig_value)
        end
        setmetatable(copy, MMY_deepCopy(getmetatable(orig)))
    else -- number, string, boolean, etc
        copy = orig
    end
    return copy
end
--------------------------------------------------------------------
function Macronomy_GetPlayerSpec()
	local _, charSpec = GetSpecializationInfo(GetSpecialization())
	if not charSpec then charSpec = "NoSpec" end
	return charSpec
end
--------------------------------------------------------------------
function Macronomy_InitNewCategory(className, specName, categoryName)
	Macronomy[MMY_locale][className][specName][categoryName] = {}
	for btnNum = 1, 12 do
		Macronomy[MMY_locale][className][specName][categoryName][MacronomyButtonList[btnNum]] = {}
		Macronomy[MMY_locale][className][specName][categoryName][MacronomyButtonList[btnNum]].btn = ""
		Macronomy[MMY_locale][className][specName][categoryName][MacronomyButtonList[btnNum]].autoShowTooltipDisabled = false
	end
	table.insert(Macronomy[MMY_locale][className][specName].categories, categoryName)
end
--------------------------------------------------------------------
function Macronomy_InitCategory(className, specName, categoryName)
	Macronomy[className][specName][categoryName] = {}
	for btnNum = 1, 12 do
		Macronomy[className][specName][categoryName][MacronomyButtonList[btnNum]] = ""
	end
	table.insert(Macronomy[className][specName].categories, categoryName)
end
--------------------------------------------------------------------
function Macronomy_SetDropDowns()
	Macronomy_setDropDownTo(MacronomyFrame_btnClass,MacronomyByChar.currentClass)
	Macronomy_setDropDownTo(MacronomyFrame_btnSpec,MacronomyByChar.currentSpec)
	Macronomy_setDropDownTo(MacronomyFrame_btnCategory,MacronomyByChar.currentCategory)
	Macronomy_setDropDownTo(MacronomyFrame_btnButton,MacronomyByChar.currentButton)
	Macronomy_setDropDownTo(MacronomyExportFrame_btnClass,MacronomyByChar.currentExportClass or "All")
	Macronomy_setDropDownTo(MacronomyExportFrame_btnSpec,MacronomyByChar.currentExportSpec or "All")
	Macronomy_setDropDownTo(MacronomyExportFrame_btnCategory,MacronomyByChar.currentExportCategory or "All")
	Macronomy_setDropDownTo(MacronomyExportFrame_btnButton,MacronomyByChar.currentExportButton or "All")

	if MacronomyByChar.currentCategory == "Default" then
		MacronomyFrame_btnCategoryRemove:Disable()
	else
		MacronomyFrame_btnCategoryRemove:Enable()
	end
	
	MacronomyMacroFrame_chkAutoTTDisable:SetChecked(Macronomy[MMY_locale][MacronomyByChar.currentClass][MacronomyByChar.currentSpec][MacronomyByChar.currentCategory][MacronomyByChar.currentButton].autoShowTooltipDisabled)
	MacronomyMacroFrame_ebxMacro:SetText(Macronomy[MMY_locale][MacronomyByChar.currentClass][MacronomyByChar.currentSpec][MacronomyByChar.currentCategory][MacronomyByChar.currentButton].btn or "")
	MacronomyMacroFrame_btnSave:Disable()
	--MacronomyMacroFrame_ebxMacro:SetFocus()
end
--------------------------------------------------------------------
function MMY_checkChangedMacro()
	if MacronomyMacroFrame_ebxMacro:GetText() == Macronomy[MMY_locale][MacronomyByChar.currentClass][MacronomyByChar.currentSpec][MacronomyByChar.currentCategory][MacronomyByChar.currentButton].btn then
		if MacronomyMacroFrame_chkAutoTTDisable:GetChecked() == Macronomy[MMY_locale][MacronomyByChar.currentClass][MacronomyByChar.currentSpec][MacronomyByChar.currentCategory][MacronomyByChar.currentButton].autoShowTooltipDisabled then
			MacronomyMacroFrame_btnSave:Disable()
		else
			MacronomyMacroFrame_btnSave:Enable()
		end
	else
		MacronomyMacroFrame_btnSave:Enable()
	end
end
--------------------------------------------------------------------
function Macronomy_escapePressed()
	Macronomy.active = false
	MacronomyFrame:Hide()
end
--------------------------------------------------------------------
function Macronomy_SetButtons()
	if MacronomyButtonUpdate and not UnitAffectingCombat("player") then
		-- Create/Set buttons to current Class/Spec/Category
		for btnNum = 1, 12 do
			local btnName = MacronomyButtonList[btnNum]
			if not GetMacroInfo(btnName) then CreateMacro(btnName, "INV_MISC_QUESTIONMARK", "#showtooltip", nil) end
			local charClass = UnitClass("player")
			local charSpec = Macronomy_GetPlayerSpec()
			local charCategory = Macronomy[MMY_locale][charClass][charSpec].categories[MacronomyByChar.categoryInUse]

			local btnText = ""
			if Macronomy.addShowTooltipByDefault and not Macronomy[MMY_locale][charClass][charSpec][charCategory][btnName].autoShowTooltipDisabled then
				btnText = btnText .. "#showtooltip\n"
			end
			btnText = btnText .. Macronomy[MMY_locale][charClass][charSpec][charCategory][btnName].btn

			btnText = MMY_replaceAliases(btnText)
			btnText = MMY_filterPreconditions(btnText)
			btnText = MMY_replaceButtonShorthandCommands(btnText)
			btnText = MMY_replaceButtonShorthandConditions(btnText)
			btnText = MMY_replaceButtonShorthandRotationResetCondtions(btnText)
			
			EditMacro(btnName,btnName, nil, btnText, 1, 1)
		end
		MacronomyButtonUpdate = false
	end
end
--------------------------------------------------------------------
function MMY_replaceAliases(txt)
	while txt:find("<a:") do
		local beforeAlias, afterAlias = MMY_splitText(txt, "<a:")
		txt = beforeAlias
		local alias, afterAliasWord = MMY_splitText(afterAlias, ">")
		local aliasReplacement, remainder = MMY_splitText(afterAliasWord, "</>")

		-- Remove newline character, if that is what the alias preceded
		if remainder:sub(1,1) == "\n" then remainder = remainder:sub(2) end
		
		txt = txt .. remainder
		txt = txt:gsub(alias, aliasReplacement)
	end

	return txt
end
--------------------------------------------------------------------
function MMY_replaceButtonShorthandCommands(txt)
	local outString = txt
	outString = outString:gsub("/c ", "/cast ")
	outString = outString:gsub("/cs ", "/castsequence ")
	return outString
end
--------------------------------------------------------------------
function MMY_filterPreconditions(txt)
	-- Evaluate and remove conditions that do not fit
	-- CONDITIONS:
	-- Race(r)
	-- Level(l)
	-- Spell(s)
	-- Item(i)

	while txt:find("</>") do
		local beforeClose, afterClose = MMY_splitText(txt, "</>")
		local reverseConditionBlock, reverseBeforeCondition = MMY_splitText(beforeClose:reverse(), "<")

		txt = reverseBeforeCondition:reverse()
		local conditionBlock = reverseConditionBlock:reverse()

		local conditionType, afterConditionType = MMY_splitText(conditionBlock, ":")
		local conditionVal, conditionalText = MMY_splitText(afterConditionType, ">")

		local myRace = UnitRace("player")
		local myLevel = UnitLevel("player")
		local isNot = conditionVal:sub(1,1) == "-"
		local isAppend = conditionVal:sub(1,1) == "+"
		conditionType = conditionType:sub(1,1):lower()
		local conditionMet = false

		local conditionMet = conditionType == "r" and conditionVal:sub(1,2):lower() == myRace:sub(1,2):lower()
		conditionMet = conditionMet or conditionType == "r" and isNot and conditionVal:sub(2,3):lower() ~= myRace:sub(1,2):lower()
		conditionMet = conditionMet or conditionType == "l" and tonumber(conditionVal) <= myLevel
		conditionMet = conditionMet or conditionType == "l" and isNot and tonumber(conditionVal:sub(2)) > myLevel
		conditionMet = conditionMet or conditionType == "s" and MMY_hasSpell(conditionVal)
		conditionMet = conditionMet or conditionType == "s" and isNot and not MMY_hasSpell(conditionVal:sub(2))
		conditionMet = conditionMet or conditionType == "s" and isAppend and MMY_hasSpell(conditionVal:sub(2))
		conditionMet = conditionMet or conditionType == "i" and MMY_hasItem(conditionVal)
		conditionMet = conditionMet or conditionType == "i" and isNot and not MMY_hasItem(conditionVal:sub(2))
		conditionMet = conditionMet or conditionType == "i" and isAppend and MMY_hasItem(conditionVal:sub(2))
		if conditionMet then
			txt = txt .. conditionalText
			if isAppend then txt = txt .. (conditionalText == "" and "/c" or "") .. " " .. conditionVal:sub(2) end
		end

		txt = txt .. afterClose
	end

	return txt
end
--------------------------------------------------------------------
function MMY_splitText(txt, delim, include)
	include = include or false
	local firstPart = ""

	if txt:find(delim) then
		firstPart = txt:sub(1, txt:find(delim) - 1)
		txt = txt:sub(txt:find(delim) + delim:len())
	else
		firstPart = txt
		txt = ""
	end

	if include == 1 then firstPart = firstPart .. delim end
	if include == 2 then txt = delim .. txt end
	return firstPart, txt
end
--------------------------------------------------------------------
function MMY_hasSpell(spellName)
	local retOK = GetSpellInfo(spellName)
	if retOK then return true end
	return false
end
--------------------------------------------------------------------
function MMY_hasItem(itemName)
	-- Check player bags for item
	local numberofslots
	local bagitemname
	for bagid=0,4 do
		numberofslots = GetContainerNumSlots(bagid)
		if (numberofslots > 0) then
			for bagslotnum=1, numberofslots do
				local bagitemid = GetContainerItemID(bagid, bagslotnum)
				if (bagitemid) then bagitemname = GetItemInfo(bagitemid) end
				if (itemName == bagitemname) then return true end
			end
		end
	end

	-- Check player equipped inventory for item
	for equipcount=1,19 do
		local equipitemid = GetInventoryItemID("player",equipcount)
		if equipitemid then equipitemname = GetItemInfo(equipitemid) end
		if itemName == equipitemname then return true end
	end
	
	return false
end
--------------------------------------------------------------------
function MMY_replaceButtonShorthandConditions(txt)
	-- Evaluate and replace shorthand for conditions
	-- * : Indicates beginning of a macro condition
	-- - : Indicates a negative condition
	-- c : Indicates ctrl key
	-- a : Indicates alt key
	-- s : Indicates shift key
	-- # : Indicates stance #
	-- i : Indicates stealth
	--
	-- One would normally only have cas, #, or i within a single or condition
	local outString = ""
			
	while txt:find("*") do
		outString = outString .. txt:sub(1, txt:find("*") - 1) .. "["
		txt = txt:sub(txt:find("*") + 1)
		local condition = txt:sub(1, txt:find("[%s*]") - 1)
		txt = txt:sub(txt:find("[%s*]") + 0)

		local resultingCondition = ""
		while condition:len() > 0 do
			local subCondition = ""

			if condition:find("&") then
				subCondition = condition:sub(1, condition:find("&") - 1)
				condition = condition:sub(condition:find("&") + 1)
			else
				subCondition = condition
				condition = ""
			end

			-- Populate pos and negative of subCondition
			local posConditions = ""
			local negConditions = ""
			if subCondition:find("-") then
				posConditions = subCondition:sub(1, subCondition:find("-") - 1)
				negConditions = subCondition:sub(subCondition:find("-") + 1)
			else
				posConditions = subCondition
			end

			-- Function to convert subCondition code to actual macro condition
			local assembledCondition = function(cond)
				local theAssembledCondition =  ""

				if cond:len() > 0 then
					if cond:sub(1,1) == "-" then theAssembledCondition = "mod"
					elseif cond:sub(1,1) == "i" then theAssembledCondition = "stealth"
					elseif tonumber(cond:sub(1,1)) ~= nil then
						theAssembledCondition = "form:"
						local stances = ""
						while cond:len() > 0 do
							if tonumber(cond:sub(1,1)) ~= nil then stances = stances .. (stances:len() > 0 and "/" or "") .. cond:sub(1,1) end
							cond = cond:sub(2)
						end
						theAssembledCondition = theAssembledCondition .. stances
					else
						theAssembledCondition = "mod:"
						local mods = ""
						while cond:len() > 0 do
							if cond:sub(1,1):lower() == "c" then mods = mods .. (mods:len() > 0 and "/" or "") .. "ctrl"
							elseif cond:sub(1,1):lower() == "a" then mods = mods .. (mods:len() > 0 and "/" or "") .. "alt"
							elseif cond:sub(1,1):lower() == "s" then mods = mods .. (mods:len() > 0 and "/" or "") .. "shift"
							end
							cond = cond:sub(2)
						end
						theAssembledCondition = theAssembledCondition .. mods
					end
				end 

				return theAssembledCondition
			end

			if posConditions:len() > 0 then resultingCondition = resultingCondition .. assembledCondition(posConditions) end
			if posConditions:len() > 0 and negConditions:len() > 0 then resultingCondition = resultingCondition .. "," end
			if negConditions:len() > 0 then resultingCondition = resultingCondition .. "no" .. assembledCondition(negConditions) end
			if condition:len() > 0 then resultingCondition = resultingCondition .. "," end
		end

		outString = outString .. resultingCondition .. "]"
	end

	outString = outString .. txt

	return outString
end
--------------------------------------------------------------------
function MMY_replaceButtonShorthandRotationResetCondtions(txt)
	while txt:find("~") do
		local outString, afterTilde = MMY_splitText(txt, "~")
		local resetConditions, afterResetConditions = MMY_splitText(afterTilde, " ", 2)

		local outResetConditions = ""
		local outResetConditionTime = ""
		while resetConditions:len() > 0 do
			if resetConditions:sub(1,1) == "t" then outResetConditions = outResetConditions .. (outResetConditions:len() > 0 and "/" or "") .. "target" end
			if resetConditions:sub(1,1) == "x" then outResetConditions = outResetConditions .. (outResetConditions:len() > 0 and "/" or "") .. "combat" end
			if resetConditions:sub(1,1) == "s" then outResetConditions = outResetConditions .. (outResetConditions:len() > 0 and "/" or "") .. "shift" end
			if resetConditions:sub(1,1) == "a" then outResetConditions = outResetConditions .. (outResetConditions:len() > 0 and "/" or "") .. "alt" end
			if resetConditions:sub(1,1) == "c" then outResetConditions = outResetConditions .. (outResetConditions:len() > 0 and "/" or "") .. "ctrl" end
			if tonumber(resetConditions:sub(1,1)) ~= nil then outResetConditionTime = outResetConditionTime .. resetConditions:sub(1,1) end

			resetConditions = resetConditions:sub(2)
		end
		if outResetConditionTime:len() > 0 then outResetConditions = outResetConditions .. (outResetConditions:len() > 0 and "/" or "") .. outResetConditionTime end
		if outResetConditions:len() > 0 then outString = outString .. "reset=" .. outResetConditions end

		txt = outString .. afterResetConditions
	end

	return txt
end
--------------------------------------------------------------------
function Macronomy_ButtonClick(btn)
	local btnName = btn:GetName()

	if btnName == "MacronomyFrame_btnX" then
		Macronomy.active = false
		MacronomyFrame:Hide()
	elseif btnName == "MacronomyFrame_btnCategoryAdd" then
		Macronomy_disableFilterSection()
		MacronomyMacroFrame:Hide()
		MacronomyNewCategoryFrame:Show()
	elseif btnName == "MacronomyFrame_btnCategoryRemove" then
		Macronomy_disableFilterSection()
		MacronomyMacroFrame:Hide()
		MacronomyDeleteCategoryFrame:Show()
		MacronomyDeleteCategoryFrame_txtConfirm:SetText("Are you sure you want to remove " .. MacronomyByChar.currentCategory .. " from " .. MacronomyByChar.currentSpec .. " " .. MacronomyByChar.currentClass)
	elseif btnName == "MacronomyMacroFrame_btnSave" then
		local classNow = MacronomyByChar.currentClass
		local specNow = MacronomyByChar.currentSpec
		local catNow = MacronomyByChar.currentCategory
		local btnNow = MacronomyByChar.currentButton
		Macronomy[MMY_locale][classNow][specNow][catNow][btnNow].btn = MacronomyMacroFrame_ebxMacro:GetText()
		Macronomy[MMY_locale][classNow][specNow][catNow][btnNow].autoShowTooltipDisabled = MacronomyMacroFrame_chkAutoTTDisable:GetChecked()
		MacronomyMacroFrame_btnSave:Disable()
		MacronomyButtonUpdate = true
	elseif btnName == "MacronomyNewCategoryFrame_btnOK" then
		Macronomy_createCategory(MacronomyNewCategoryFrame_ebxNewCat:GetText())
		Macronomy_enableFilterSection()
		MacronomyNewCategoryFrame_ebxNewCat:SetText("")
		MacronomyMacroFrame:Show()
		MacronomyNewCategoryFrame:Hide()
	elseif btnName == "MacronomyNewCategoryFrame_btnCancel" then
		Macronomy_enableFilterSection()
		MacronomyNewCategoryFrame_ebxNewCat:SetText("")
		MacronomyMacroFrame:Show()
		MacronomyNewCategoryFrame:Hide()
	elseif btnName == "MacronomyDeleteCategoryFrame_btnYes" then
		Macronomy_enableFilterSection()
		Macronomy_removeCategory()
		MacronomyMacroFrame:Show()
		MacronomyDeleteCategoryFrame:Hide()
	elseif btnName == "MacronomyDeleteCategoryFrame_btnNo" then
		Macronomy_enableFilterSection()
		MacronomyMacroFrame:Show()
		MacronomyDeleteCategoryFrame:Hide()
	elseif btnName == "MacronomyMacroFrame_btnImport" then
		MacronomyMacroFrame:Hide()
		MacronomyImportFrame_scrollFrame_ebxImport:SetText("")
		MacronomyImportFrame:Show()
		Macronomy_disableFilterSection()
	elseif btnName == "MacronomyImportFrame_btnImport" then
		Macronomy_ImportText()
		MacronomyMacroFrame:Show()
		MacronomyImportFrame:Hide()
		Macronomy_enableFilterSection()
	elseif btnName == "MacronomyImportFrame_btnCancel" then
		MacronomyMacroFrame:Show()
		MacronomyImportFrame:Hide()
		Macronomy_enableFilterSection()
	elseif btnName == "MacronomyMacroFrame_btnExport" then
		MacronomyByChar.currentExportClass = "All"
		MacronomyByChar.currentExportSpec = "All"
		MacronomyByChar.currentExportCategory = "All"
		MacronomyByChar.currentExportButton = "All"
		Macronomy_UpdateExportText()
		Macronomy_SetDropDowns()
		MacronomyMacroFrame:Hide()
		MacronomyExportFrame:Show()
		Macronomy_disableFilterSection()
	elseif btnName == "MacronomyMacroFrame_btnOptions" then
		MacronomyMacroFrame:Hide()
		MacronomyOptionsFrame:Show()
		Macronomy_disableFilterSection()
	elseif btnName == "MacronomyOptionsFrame_btnOK" then
		MacronomyMacroFrame:Show()
		MacronomyOptionsFrame:Hide()
		Macronomy_enableFilterSection()
	elseif btnName == "MacronomyExportFrame_btnOK" then
		MacronomyMacroFrame:Show()
		MacronomyExportFrame:Hide()
		Macronomy_enableFilterSection()
	end
end
--------------------------------------------------------------------
function MMY_chkClick(chk)
	local chkName = chk:GetName()

	if chkName == "MacronomyOptionsFrame_chkShowTTByDefault" then
		Macronomy.addShowTooltipByDefault = MacronomyOptionsFrame_chkShowTTByDefault:GetChecked()
		MMY_setDisableTTVisibleState()
		MacronomyButtonUpdate = true
	end
end
--------------------------------------------------------------------
function MMY_setDisableTTVisibleState()
	if Macronomy.addShowTooltipByDefault then
		MacronomyMacroFrame_chkAutoTTDisable:Show()
	else
		MacronomyMacroFrame_chkAutoTTDisable:Hide()
	end
end
--------------------------------------------------------------------
local MacronomyDropDownSize = {}
MacronomyDropDownSize["MacronomyFrame_btnClass"] = 120
MacronomyDropDownSize["MacronomyFrame_btnSpec"] = 120
MacronomyDropDownSize["MacronomyFrame_btnCategory"] = 120
MacronomyDropDownSize["MacronomyFrame_btnButton"] = 50
MacronomyDropDownSize["MacronomyExportFrame_btnClass"] = 120
MacronomyDropDownSize["MacronomyExportFrame_btnSpec"] = 120
MacronomyDropDownSize["MacronomyExportFrame_btnCategory"] = 120
MacronomyDropDownSize["MacronomyExportFrame_btnButton"] = 50

function Macronomy_dropMenuShow(self)
	UIDropDownMenu_Initialize(self, Macronomy_dropMenuInitialize);
	UIDropDownMenu_SetWidth(self,MacronomyDropDownSize[self:GetName()],3);
	UIDropDownMenu_JustifyText(self, "LEFT");
end

function Macronomy_dropMenuInitialize(btn)
	local listForBtn
	if btn:GetName() == "MacronomyFrame_btnClass" then
		listForBtn = MacronomyClassList
	elseif btn:GetName() == "MacronomyExportFrame_btnClass" then
		listForBtn = MacronomyExportClassList
	elseif btn:GetName() == "MacronomyFrame_btnSpec" then
		local currentClass = UnitClass("player")
		if MacronomyByChar then currentClass = MacronomyByChar.currentClass end
		listForBtn = MacronomySpecList[currentClass]
	elseif btn:GetName() == "MacronomyExportFrame_btnSpec" then
		local currentClass = "All"
		if MacronomyByChar then currentClass = MacronomyByChar.currentExportClass or "All" end
		listForBtn = MacronomyExportSpecList[currentClass]
	elseif btn:GetName() == "MacronomyFrame_btnCategory" then
		listForBtn = {"Default"}
		if Macronomy and MacronomyByChar then listForBtn = Macronomy[MMY_locale][MacronomyByChar.currentClass][MacronomyByChar.currentSpec].categories end
	elseif btn:GetName() == "MacronomyExportFrame_btnCategory" then
		listForBtn = {"All"}
		if MacronomyByChar and MacronomyByChar.currentExportSpec ~= "All" then
			for catCount=1,getn(Macronomy[MMY_locale][MacronomyByChar.currentExportClass][MacronomyByChar.currentExportSpec].categories) do
				listForBtn[catCount + 1] = Macronomy[MMY_locale][MacronomyByChar.currentExportClass][MacronomyByChar.currentExportSpec].categories[catCount]
			end
		end
	elseif btn:GetName() == "MacronomyFrame_btnButton" then
		listForBtn = MacronomyButtonList
	elseif btn:GetName() == "MacronomyExportFrame_btnButton" then
		listForBtn = MacronomyExportButtonList
		if MacronomyByChar and MacronomyByChar.currentExportCategory == "All" then listForBtn = {"All"} end
	end

	Macronomy_populateButton(btn, listForBtn)
end

function Macronomy_populateButton(btn, listForBtn)
	if listForBtn then
		for listCount=1,getn(listForBtn) do
			Macronomy_addDropDownMenuBtn(listForBtn[listCount], btn)
		end
	end
end

function Macronomy_addDropDownMenuBtn(str, btn)
	local info = UIDropDownMenu_CreateInfo()
	info.text = str
	info.func = Macronomy_dropDownMenuBtnClick
	info.owner = btn
	UIDropDownMenu_AddButton(info)
end

function Macronomy_dropDownMenuBtnClick(self)
	local filtersChanged = false
	local exportFiltersChanged = false
	local parentFrame = self.owner;
	local selection = self:GetText();
	Macronomy_setDropDownTo(parentFrame, selection)
	
	if parentFrame == MacronomyFrame_btnClass and MacronomyByChar.currentClass ~= selection then
		MacronomyByChar.currentClass = selection
		MacronomyByChar.currentSpec = MacronomySpecList[MacronomyByChar.currentClass][1]
		if MacronomyByChar.currentClass == UnitClass("player") then
			MacronomyByChar.currentSpec = Macronomy_GetPlayerSpec()
		end
		MacronomyByChar.currentCategory = "Default"
		MacronomyByChar.currentButton = "m1"
		Macronomy_dropMenuInitialize(MacronomyFrame_btnClass)
		filtersChanged = true
	end

	if parentFrame == MacronomyFrame_btnSpec and MacronomyByChar.currentSpec ~= selection then
		MacronomyByChar.currentSpec = selection
		MacronomyByChar.currentCategory = "Default"
		MacronomyByChar.currentButton = "m1"
		filtersChanged = true
	end
	
	if parentFrame == MacronomyFrame_btnCategory and MacronomyByChar.currentCategory ~= selection then
		MacronomyByChar.currentCategory = selection
		filtersChanged = true
	end

	if parentFrame == MacronomyFrame_btnButton and MacronomyByChar.currentButton ~= selection then
		MacronomyByChar.currentButton = selection
		filtersChanged = true
	end
	
	if parentFrame == MacronomyExportFrame_btnClass and MacronomyByChar.currentExportClass ~= selection then
		MacronomyByChar.currentExportClass = selection
		MacronomyByChar.currentExportSpec = "All"
		MacronomyByChar.currentExportCategory = "All"
		MacronomyByChar.currentExportButton = "All"
		exportFiltersChanged = true
	end

	if parentFrame == MacronomyExportFrame_btnSpec and MacronomyByChar.currentExportSpec ~= selection then
		MacronomyByChar.currentExportSpec = selection
		MacronomyByChar.currentExportCategory = "All"
		MacronomyByChar.currentExportButton = "All"
		exportFiltersChanged = true
	end

	if parentFrame == MacronomyExportFrame_btnCategory and MacronomyByChar.currentExportCategory ~= selection then
		MacronomyByChar.currentExportCategory = selection
		MacronomyByChar.currentExportButton = "All"
		exportFiltersChanged = true
	end

	if parentFrame == MacronomyExportFrame_btnButton and MacronomyByChar.currentExportButton ~= selection then
		MacronomyByChar.currentExportButton = selection
		exportFiltersChanged = true
	end
	
	if exportFiltersChanged then Macronomy_UpdateExportText() end

	if filtersChanged or exportFiltersChanged then
		Macronomy_SetDropDowns()
	end
end

function Macronomy_setDropDownTo(btn, str)
	UIDropDownMenu_SetSelectedValue(btn, str)
	UIDropDownMenu_SetText(btn, str)
end
--------------------------------------------------------------------
function Macronomy_createCategory(newCat)
	local createNewCategory = true
	if newCat == "" or newCat == "Default" then createNewCategory = false end
	for catCount=1,getn(Macronomy[MMY_locale][MacronomyByChar.currentClass][MacronomyByChar.currentSpec].categories) do
		if newCat == Macronomy[MMY_locale][MacronomyByChar.currentClass][MacronomyByChar.currentSpec].categories[catCount] then createNewCategory = false end
	end

	if createNewCategory then
		Macronomy[MMY_locale][MacronomyByChar.currentClass][MacronomyByChar.currentSpec].categories[getn(Macronomy[MMY_locale][MacronomyByChar.currentClass][MacronomyByChar.currentSpec].categories) + 1] = newCat
		Macronomy_InitCategory(MacronomyByChar.currentClass, MacronomyByChar.currentSpec, newCat)
		MacronomyByChar.currentCategory = newCat

		Macronomy_SetDropDowns()
	end
end
--------------------------------------------------------------------
function Macronomy_shiftCategory()
	MacronomyByChar.categoryInUse = MacronomyByChar.categoryInUse + 1
	if MacronomyByChar.categoryInUse > getn(Macronomy[MMY_locale][UnitClass("player")][Macronomy_GetPlayerSpec()].categories) then MacronomyByChar.categoryInUse = 1 end
	MacronomyButtonUpdate = true
	MMY_echo("Macros set to category: " .. Macronomy_GetPlayerSpec() .. " " .. UnitClass("player") .. " " .. Macronomy[MMY_locale][UnitClass("player")][Macronomy_GetPlayerSpec()].categories[MacronomyByChar.categoryInUse], true)
end
--------------------------------------------------------------------
function Macronomy_removeCategory()
	if MacronomyByChar.currentCategory ~= "Default" then
		local isPastCategory = false
		local isCurrentRole = false
		local catTotalCount = getn(Macronomy[MMY_locale][MacronomyByChar.currentClass][MacronomyByChar.currentSpec].categories)
		for catCount=1,catTotalCount do
			if isPastCategory then
				Macronomy[MMY_locale][MacronomyByChar.currentClass][MacronomyByChar.currentSpec].categories[catCount - 1] = Macronomy[MMY_locale][MacronomyByChar.currentClass][MacronomyByChar.currentSpec].categories[catCount]
			elseif MacronomyByChar.currentCategory == Macronomy[MMY_locale][MacronomyByChar.currentClass][MacronomyByChar.currentSpec].categories[catCount] then
				if MacronomyByChar.currentClass == UnitClass("player") and MacronomyByChar.currentSpec == Macronomy_GetPlayerSpec() and MacronomyByChar.categoryInUse == catCount then
					MacronomyByChar.categoryInUse = 1
					MacronomyButtonUpdate = true
				end
				isPastCategory = true
			end
		end
		Macronomy[MMY_locale][MacronomyByChar.currentClass][MacronomyByChar.currentSpec].categories[catTotalCount] = nil
		Macronomy[MMY_locale][MacronomyByChar.currentClass][MacronomyByChar.currentSpec][MacronomyByChar.currentCategory] = nil
		MacronomyByChar.currentCategory = "Default"
		Macronomy_SetDropDowns()
	end
end
--------------------------------------------------------------------
function Macronomy_UpdateExportText()
	local exportText = ""

	for exportClassCount = 1, getn(MacronomyClassList) do
		if MacronomyByChar.currentExportClass == "All" or MacronomyByChar.currentExportClass == MacronomyClassList[exportClassCount] then
			local exportLoopClass = MacronomyClassList[exportClassCount]
			exportText = exportText .. (exportText ~= "" and "~~~~~" or "") .. exportLoopClass .. ":"

			local numberOfExportSpecs = 0
			for exportSpecCount = 1, getn(MacronomySpecList[MacronomyClassList[exportClassCount]]) do
				if MacronomyByChar.currentExportSpec == "All" or MacronomyByChar.currentExportSpec == MacronomySpecList[exportLoopClass][exportSpecCount] then
					local exportLoopSpec = MacronomySpecList[exportLoopClass][exportSpecCount]
					numberOfExportSpecs = numberOfExportSpecs + 1
					exportText = exportText .. (numberOfExportSpecs > 1 and "~~~~" or "") .. exportLoopSpec .. ":"
				
					local numberOfExportCategories = 0
					for exportCategoryCount = 1, getn(Macronomy[MMY_locale][MacronomyClassList[exportClassCount]][exportLoopSpec].categories) do
						if MacronomyByChar.currentExportCategory == "All" or MacronomyByChar.currentExportCategory == Macronomy[MMY_locale][exportLoopClass][exportLoopSpec].categories[exportCategoryCount] then
							local exportLoopCategory = Macronomy[MMY_locale][exportLoopClass][exportLoopSpec].categories[exportCategoryCount]
							numberOfExportCategories = numberOfExportCategories + 1
							exportText = exportText .. (numberOfExportCategories > 1 and "~~~" or "") .. exportLoopCategory .. ":"

							local numberOfExportButtons = 0
							for exportButtonCount = 1, getn(MacronomyButtonList) do
								if MacronomyByChar.currentExportButton == "All" or MacronomyByChar.currentExportButton == MacronomyButtonList[exportButtonCount] then
									local exportLoopButton = MacronomyButtonList[exportButtonCount]
									local btnText = string.gsub(Macronomy[MMY_locale][exportLoopClass][exportLoopSpec][exportLoopCategory][exportLoopButton].btn, "\n", "/n/")
									numberOfExportButtons = numberOfExportButtons + 1
									exportText = exportText .. (numberOfExportButtons > 1 and "~~" or "") .. exportLoopButton .. btnText
								end
							end
						end
					end
				end
			end
		end
	end
	
	MacronomyExportFrame_scrollFrame_ebxExport:SetText(exportText)
end
--------------------------------------------------------------------
function MMY_importTextChanged()
	if MacronomyImportFrame_scrollFrame_ebxImport:GetText():len() > 0 then
		MacronomyImportFrame_btnImport:Enable()
	else
		MacronomyImportFrame_btnImport:Disable()
	end
end
--------------------------------------------------------------------
function Macronomy_ImportText()
	local importText = MacronomyImportFrame_scrollFrame_ebxImport:GetText()
	if importText ~= "" then
		local importClassArray = Macronomy_stringSplit(importText, "~~~~~")
		for importClassCount = 1, getn(importClassArray) do
			local classToImport, strAfterClass = Macronomy_oneWordSplit(importClassArray[importClassCount], ":")
			local importSpecArray = Macronomy_stringSplit(strAfterClass, "~~~~")
			for importSpecCount = 1, getn(importSpecArray) do
				local specToImport, strAfterSpec = Macronomy_oneWordSplit(importSpecArray[importSpecCount], ":")
				local importCategoryArray = Macronomy_stringSplit(strAfterSpec, "~~~")
				for importCategoryCount = 1, getn(importCategoryArray) do
					local categoryToImport, strAfterCategory = Macronomy_oneWordSplit(importCategoryArray[importCategoryCount], ":")
					if not Macronomy[MMY_locale][classToImport][specToImport][categoryToImport] then Macronomy_InitCategory(classToImport, specToImport, categoryToImport) end
					local importButtonArray = Macronomy_stringSplit(strAfterCategory, "~~")
					for importButtonCount = 1, getn(importButtonArray) do
						local importButtonName = string.sub(importButtonArray[importButtonCount],1,2)
						local importButtonContent = string.gsub(string.sub(importButtonArray[importButtonCount],3), "/n/", "\n")
						Macronomy[MMY_locale][classToImport][specToImport][categoryToImport][importButtonName].btn = importButtonContent
					end
				end
			end
		end
		
		MacronomyButtonUpdate = true
		MacronomyMacroFrame_ebxMacro:SetText(Macronomy[MMY_locale][MacronomyByChar.currentClass][MacronomyByChar.currentSpec][MacronomyByChar.currentCategory][MacronomyByChar.currentButton].btn or "")
	end
end
--------------------------------------------------------------------
function Macronomy_stringSplit(str, delim)
	local outputArray = {}

	while str ~= "" do
		local delimStart = string.find(str, delim)
		outputArray[getn(outputArray) + 1] = (delimStart ~= nil and string.sub(str,1,delimStart - 1) or str)
		str = (delimStart ~= nil and string.sub(str,delimStart + string.len(delim)) or "")
	end
	
	return outputArray
end
--------------------------------------------------------------------
function Macronomy_oneWordSplit(str, delim)
	local delimLoc = string.find(str, delim)
	local rtnStr = (delimLoc ~= nil and string.sub(str, 1, delimLoc - 1) or str)
	local rtnRemainStr = (delimLoc ~= nil and string.sub(str, delimLoc + 1) or "")
	return rtnStr, rtnRemainStr
end
--------------------------------------------------------------------
function Macronomy_disableFilterSection()
	MacronomyFrame_btnCategoryAdd:Disable()
	MacronomyFrame_btnCategoryRemove:Disable()
	UIDropDownMenu_DisableDropDown(MacronomyFrame_btnClass)
	UIDropDownMenu_DisableDropDown(MacronomyFrame_btnSpec)
	UIDropDownMenu_DisableDropDown(MacronomyFrame_btnCategory)
	UIDropDownMenu_DisableDropDown(MacronomyFrame_btnButton)
end
--------------------------------------------------------------------
function Macronomy_enableFilterSection()
	MacronomyFrame_btnCategoryAdd:Enable()
	MacronomyFrame_btnCategoryRemove:Enable()
	UIDropDownMenu_EnableDropDown(MacronomyFrame_btnClass)
	UIDropDownMenu_EnableDropDown(MacronomyFrame_btnSpec)
	UIDropDownMenu_EnableDropDown(MacronomyFrame_btnCategory)
	UIDropDownMenu_EnableDropDown(MacronomyFrame_btnButton)
end
--------------------------------------------------------------------
function MMY_scrollFrameClick()
	MacronomyMacroFrame_ebxMacro:SetCursorPosition(MacronomyMacroFrame_ebxMacro:GetNumLetters())
	MacronomyMacroFrame_ebxMacro:SetFocus()
end
--------------------------------------------------------------------
SLASH_MACRONOMYSHIFT1 = "/macronomyshift"
function SlashCmdList.MACRONOMYSHIFT(msg, editbox)
	Macronomy_shiftCategory()
end

SLASH_MACRONOMYSHOW1 = "/macronomy"
SLASH_MACRONOMYSHOW2 = "/mmy"
function SlashCmdList.MACRONOMYSHOW(msg, editbox)
	if Macronomy.active then
		MacronomyFrame:Hide()
	else
		MacronomyFrame:Show()
		Macronomy_SetDropDowns()
	end
	Macronomy.active = not Macronomy.active
end

SLASH_MACRONOMYREVERT1 = "/mmyrevert"
function SlashCmdList.MACRONOMYREVERT(msg, editbox)
	if Macronomy.enUS ~= nil then
		Macronomy.enUS = nil
	end
end

--------------------------------------------------------------------

local Macronomy_spellClicked = function(button)
	if IsShiftKeyDown() and MacronomyMacroFrame_ebxMacro:HasFocus() then
		local btnNum = button:GetID() <= 6 and ((button:GetID() * 2) - 1) or ((button:GetID() - 6) * 2)
		MacronomyMacroFrame_ebxMacro:Insert(_G["SpellButton" .. btnNum .. "SpellName"]:GetText())
	end
end
hooksecurefunc("SpellButton_OnModifiedClick", Macronomy_spellClicked)

local Macronomy_equipClicked = function(button)
	if IsShiftKeyDown() and MacronomyMacroFrame_ebxMacro:HasFocus() then
		local itemId = GetInventoryItemID("player", GetInventorySlotInfo(string.sub(button:GetName(),10)))
		if itemId then
			local itemName = GetItemInfo(itemId)
			MacronomyMacroFrame_ebxMacro:Insert(itemName)
		end
	end
end
hooksecurefunc("PaperDollItemSlotButton_OnModifiedClick", Macronomy_equipClicked)

local Macronomy_bagItemClicked = function(button)
	if IsShiftKeyDown() and MacronomyMacroFrame_ebxMacro:HasFocus() then
		local btnName = button:GetName()
		local bagViewer = (string.find(btnName, "ARKINV_Frame") and "ArkInv" or "Default")
		local indexOfItem = string.find(btnName, "Item")
		local bagNum = tonumber(string.sub(btnName, indexOfItem - 1, indexOfItem - 1)) - 1
		local slotNum = tonumber(string.sub(btnName, indexOfItem + 4))
		
		local itemId = GetContainerItemID(bagNum, slotNum)
		if itemId then
			local itemName = GetItemInfo(itemId)
			MacronomyMacroFrame_ebxMacro:Insert(itemName)
		end
	end
end
hooksecurefunc("ContainerFrameItemButton_OnModifiedClick", Macronomy_bagItemClicked)

local MMY_offClick = function(self,button)
	if MacronomyMacroFrame_ebxMacro:HasFocus() then 
		MacronomyMacroFrame_ebxMacro:ClearFocus()
	end
end
WorldFrame:HookScript("OnMouseUp",MMY_offClick)